#include <QtGui>
#include "principal.h"


// Recupera_op recupera uno de los tres argumentos de una instruccion. En el caso de ARG_ENTERO, ARG_ETIQUETA o
// ARG_REAL recuperar el valor es inmediato. Si el argumento es ARG_POSICION debera leer el dato de la memoria.
// Leer de un argumento ARG_NULO dara un ERROR_TIPO.
// Los metodos recupera_op estan sobrecargados. Los parametros que admiten son:
//
// IN c_tipo_ctd *inst -> Un puntero a una instancia de la clase instruccion de la cual queremos obtener un argumento.
// IN int narg -> Uno de los tres argumentos que necesitamos de la instruccion. Numerados de cero a dos.
// OUT int *val -> Salida por referencia del entero que buscamos.
// OUT float *fval -> Salida por referencia del real que bucamos.
// IN tipo_mem tipo_dato -> Especificamos si el entero que buscamos es:
//                          MEM_INT: Un entero. Es el valor por defecto en caso de no especificarlo.
//                          MEM_FP: Un puntero a marco (escrito en memoria por PUSHFP)
//                          MEM_DEST_S: Una direccion de destino de salto (escrito en memoria por CALL)
//                          MEM_UNUSED: No queremos que busque en memoria.
//                          No deberiamos usar MEM_REAL, para ello usar el metodo sobrecargado adecuado.
//
// Deberiamos comprobar la salida en busca de posibles errores. Los errores pueden ser:
//
//                          NO_ERROR: No se han producido errores. Equivale al valor 0.
//                          ERROR_VACIA: Se ha ido a buscar el dato a memoria y la celda de memoria estaba vacia.
//                          ERROR_NOEXISTE: Se ha ido a buscar el dato a memoria (El argumento era ARG_POSICION) pero
//                                          la direccion era erronea (negativa o superior a MAXMEM)
//                          ERROR_TIPO: Se ha ido a buscar el dato a memoria pero el dato alli contenido no coincidia con
//                                      el tipo de dato buscado.
//                          ERROR_ARG: El valor de entrada narg no es valido (debe ser un entero entre cero y dos)
//                          ERROR_SUP_TOP: Se ha buscado en memoria el dato y la direccion de memoria era superior al
//                                         puntero a cima.

error_mem Principal::recupera_op(int narg, c_tipo_ctd *inst, int *val, tipo_mem tipo_dato){
    t_cod_arg tipo;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo == ARG_ENTERO) return inst->r_int(narg, val);
    if(tipo == ARG_ETIQUETA) return inst->r_eti(narg, val);
    if(tipo == ARG_POSICION) return leer_mem(narg, inst, val, tipo_dato);
    return ERROR_TIPO;
}

error_mem Principal::recupera_op(int narg, c_tipo_ctd *inst, float *fval){
    t_cod_arg tipo;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo == ARG_REAL) return inst->r_rea(narg, fval);
    if(tipo == ARG_POSICION) return leer_mem(narg, inst, fval);
    return ERROR_TIPO;
}

error_mem Principal::recupera_op(int narg, c_tipo_ctd *inst, bool *niv, int *pos){
    t_cod_arg tipo;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo == ARG_POSICION) return inst->r_pos(narg, niv, pos); // Devuelve por referencia
    return ERROR_TIPO;
}


// Escribe_mem escribe un dato entero o real en memoria.
// El metodo esta sobrecargado. Los parametros que admite son:
//
// IN c_tipo_ctd *inst -> Un puntero a una instancia de la clase instruccion. Nos interesara
//                        si uno de sus argumentos es ARG_POSICION, ya que sera la posicion de memoria donde
//                        se escribira.
// IN int narg -> Senyala el argumento que contiene la direccion de memoria (ARG_POSICION) donde queremos guardar el dato,
//                numerado de cero a dos.
// IN int val -> Dato entero que queremos guardar en memoria.
// IN float fval -> Dato real que queremos guardar en memoria.
// IN bool p_muestra -> Indicamos a la funcion si queremos que los cambios en memoria se reflejen en la representacion
//                      grafica del programa. Si estamos ejecutando el programa paso a paso debera ser TRUE. En caso de
//                      ejecucion continua deberia ser FALSE.
// IN tipo_mem tipo_dato -> Especificamos que el entero que guardamos es:
//                          MEM_INT: Un entero
//                          MEM_FP: Un puntero a marco (escrito en memoria por PUSHFP)
//                          MEM_DEST_S: Una direccion de destino de salto (escrito en memoria por CALL)
//                          Los demas casos de la enumeracion no deberian usarse.
//                          Por defecto tipo_dato es MEM_INT.
// IN int de_inst -> Especificamos el numero de instruccion que guarda el dato. El argumento es solo relevante cuando
//                   guardamos en memoria una puntero a cima o a marco (MEM_FP o MEM_DEST_S). En otros casos el valor
//                   se ignora. Por defecto el valor es cero.
// IN bool niv -> El nivel de la memoria (local o global) en caso que guardemos un dato en memoria indicando nivel y posicion.
// IN int posi -> La posicion en memoria en caso que guardemos un dato en memoria indicando nivel y posicion. La direccion
//                efectiva se calcula como (niv ? p_fp + pos : pos)
// IN int pos -> La direccion de memoria fisica, en caso que queramos indicar la direccion de memoria de forma absoluta,
//               indicando un numero entre 0 y MAXMEM.
//
// Los metodos pueden leer o escriben en las siguientes variables de clase:
//
//      IN Principal::pila_estricta: Si pila_estricta es TRUE, cualquier acceso a memoria por encima del puntero a cima no
//                                   sera valido y los tipos MEM_INT, MEM_FP y MEM_DEST_S no seran equivalentes.
//      IN Principal::p_top: Lee la posicion del puntero a cima para comprobar que se escribe por debajo de su valor.
//      IN Principal::p_fp: Lee la posicion del puntero a marco.
//      OUT Principal::ult_escr: Escribe la direccion de memoria fisica de la ultima escritura (util para representar la ultima
//                               posicion de memoria escrita con un sombreado).
//      OUT Principal::list_mem: Modifica la memoria al escribir el dato.
//      OUT Principal::rep_mem: Si p_muestra es true, modifica la representacion grafica de la memoria al escribir el dato.
//
// Deberiamos comprobar la salida en busca de posibles errores. Los errores pueden ser:
//
//                          NO_ERROR: No se han producido errores. Equivale al valor 0.
//                          ERROR_ARG: El valor de entrada narg no es valido (debe ser un entero entre cero y dos)
//                          ERROR_SUP_TOP: Se ha buscado en memoria el dato y la direccion de memoria era superior al
//                                         puntero a cima.

error_mem Principal::escribe_mem(int narg, c_tipo_ctd *inst, int val, bool p_muestra, tipo_mem tipo_dato, int de_inst){
    t_cod_arg tipo;
    bool n;
    int p;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo != ARG_POSICION) return ERROR_TIPO;
    err = inst->r_pos(narg, &n, &p);
    if(err != NO_ERROR) return err;
    return escribe_mem(n, p, val, p_muestra, tipo_dato, de_inst);
}

error_mem Principal::escribe_mem(int narg, c_tipo_ctd *inst, float fval, bool p_muestra){
    t_cod_arg tipo;
    bool n;
    int p;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo != ARG_POSICION) return ERROR_TIPO;
    err = inst->r_pos(narg, &n, &p);
    if(err != NO_ERROR) return err;
    return escribe_mem(n, p, fval, p_muestra);
}

error_mem Principal::escribe_mem(bool niv, int posi, int val, bool p_muestra, tipo_mem tipo_dato, int de_inst){
    if(niv) return escribe_mem(p_fp + posi, val, p_muestra, tipo_dato, de_inst);
    return escribe_mem(posi, val, p_muestra, tipo_dato, de_inst);
}

error_mem Principal::escribe_mem(bool niv, int posi, float fval, bool p_muestra){
    if(niv) return escribe_mem(p_fp + posi, fval, p_muestra);
    return escribe_mem(posi, fval, p_muestra);
}

error_mem Principal::escribe_mem(int pos, int val, bool p_muestra, tipo_mem tipo_dato, int de_inst){
    if(pos < 0 || pos >= MAXMEM) return ERROR_NOEXISTE;
    if(pos >= p_top && pila_estricta) return ERROR_SUP_TOP;
    switch(tipo_dato){
      case MEM_INT:
	list_mem.at(pos)->s_int(val); break;
      case MEM_FP:
	list_mem.at(pos)->s_fpo(val, de_inst); break;
      case MEM_DEST_S:
	list_mem.at(pos)->s_des(val, de_inst); break;
      default:
	break;
    }
    if(p_muestra){
	switch(tipo_dato){
	  case MEM_INT:
	    rep_mem.at(pos)->setText(3, QString("E:"));
	    if(val >= 32 && val <= 126) rep_mem.at(pos)->setText(5, QString(QChar(val)));
	    else rep_mem.at(pos)->setText(5, QString());
	    break;
	  case MEM_FP:
	    rep_mem.at(pos)->setText(3, QString("FP (%1):").arg(de_inst));
	    break;
	  case MEM_DEST_S:
	    rep_mem.at(pos)->setText(3, QString("DIR (%1):").arg(de_inst));
	    break;
	  default:
	    break;
	}
	rep_mem.at(pos)->setText(4, QString("%1").arg(val));
	for(int i = 0; i < 6; i++) rep_mem.at(pos)->setBackground(i, QBrush(Qt::Dense6Pattern));
	if(desp_mem) tree_mem->scrollToItem(rep_mem.at(pos));
    }
    ult_escr = pos;
    return NO_ERROR;
}

error_mem Principal::escribe_mem(int pos, float fval, bool p_muestra){
    if(pos < 0 || pos >= MAXMEM) return ERROR_NOEXISTE;
    if(pos >= p_top && pila_estricta) return ERROR_SUP_TOP;
    list_mem.at(pos)->s_rea(fval);
    if(p_muestra){
	rep_mem.at(pos)->setText(3, QString("R:"));
	rep_mem.at(pos)->setText(4, QString("%1").arg(fval));
	rep_mem.at(pos)->setText(5, QString());
	for(int i = 0; i < 6; i++) rep_mem.at(pos)->setBackground(i, QBrush(Qt::Dense6Pattern));
	if(desp_mem) tree_mem->scrollToItem(rep_mem.at(pos));
    }
    ult_escr = pos;
    return NO_ERROR;
}


// Leer_mem lee un dato entero o real en memoria.
// El metodo esta sobrecargado. Los parametros que admite son:
//
// IN c_tipo_ctd *inst -> Un puntero a una instancia de la clase instruccion. Nos interesara
//                        si uno de sus argumentos es ARG_POSICION, ya que sera la posicion de memoria de donde
//                        se leera.
// IN int narg -> Senyala el argumento que contiene la direccion de memoria (ARG_POSICION) de donde queremos leer el dato,
//                numerado de cero a dos.
// OUT int *val -> Salida por referencia del dato entero que estamos leyendo.
// OUT float *fval -> Salida por referencia del dato real que estamos leyendo.
// IN tipo_mem tipo_dato -> Especificamos que el entero que queremos leer es:
//                          MEM_INT: Un entero
//                          MEM_FP: Un puntero a marco (escrito en memoria por PUSHFP)
//                          MEM_DEST_S: Una direccion de destino de salto (escrito en memoria por CALL)
//                          Los demas casos de la enumeracion no deberian usarse.
//                          Por defecto tipo_dato es MEM_INT.
// IN bool niv -> El nivel de la memoria (local o global) en caso que leamos un dato en memoria indicando nivel y posicion.
// IN int posi -> La posicion en memoria en caso que leamos un dato en memoria indicando nivel y posicion. La direccion
//                efectiva se calcula como (niv ? p_fp + pos : pos)
// IN int pos -> La direccion de memoria fisica, en caso de que pasemos la direccion de memoria de forma absoluta,
//               indicando un numero entre 0 y MAXMEM.
//
// Los metodos pueden leer o escriben en las siguientes variables de clase:
//
//      IN Principal::pila_estricta: Si pila_estricta es TRUE, cualquier acceso a memoria por encima del puntero a cima no
//                                   sera valido y los tipos MEM_INT, MEM_FP y MEM_DEST_S no seran equivalentes.
//      IN Principal::p_top: Lee la posicion del puntero a cima para comprobar que se lee por debajo de su valor.
//      IN Principal::p_fp: Lee la posicion del puntero a marco.
//      IN Principal::list_mem: Lee el dato de la memoria.
//
// Deberiamos comprobar la salida en busca de posibles errores. Los errores pueden ser:
//
//                          NO_ERROR: No se han producido errores. Equivale al valor 0.
//                          ERROR_ARG: El valor de entrada narg no es valido (debe ser un entero entre cero y dos).
//                          ERROR_TIPO: El dato en memoria no era del mismo tipo que el dato buscado.
//                          ERROR_SUP_TOP: Se ha buscado en memoria el dato y la direccion de memoria era superior al
//                                         puntero a cima.
//                          ERROR_VACIA: Se lee de una posicion de memoria que esta vacia.
//                          ERROR_NOEXISTE: Se lee de una direccion fisica de memoria que no existe (por debajo de cero
//                                          o superior a MAXMEM)

error_mem Principal::leer_mem(int narg, c_tipo_ctd *inst, int *val, tipo_mem tipo_dato){
    t_cod_arg tipo;
    int pos;
    bool niv;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo != ARG_POSICION) return ERROR_TIPO;
    err = inst->r_pos(narg, &niv, &pos);
    if(err != NO_ERROR) return err;
    return leer_mem(niv, pos, val, tipo_dato);
}

error_mem Principal::leer_mem(int narg, c_tipo_ctd *inst, float *fval){
    t_cod_arg tipo;
    int pos;
    bool niv;
    error_mem err = inst->r_arg(narg, &tipo);
    if(err != NO_ERROR) return err;
    if(tipo != ARG_POSICION) return ERROR_TIPO;
    err = inst->r_pos(narg, &niv, &pos);
    if(err != NO_ERROR) return err;
    return leer_mem(niv, pos, fval);
}

error_mem Principal::leer_mem(bool niv, int pos, int *val, tipo_mem tipo_dato){
    if(niv) return leer_mem(p_fp + pos, val, tipo_dato);
    return leer_mem(pos, val, tipo_dato);
}

error_mem Principal::leer_mem(bool niv, int pos, float *fval){
    if(niv) return leer_mem(p_fp + pos, fval);
    return leer_mem(pos, fval);
}

error_mem Principal::leer_mem(int pos, int *val, tipo_mem tipo_dato){
    if(pos < 0 || pos >= MAXMEM) return ERROR_NOEXISTE;
    /* Falta devolver error por encima de pila */
    if(pos >= p_top && pila_estricta) return ERROR_SUP_TOP;
    int trash;
    if(pila_estricta) switch(tipo_dato){
      case MEM_INT:
	return list_mem.at(pos)->r_int(val);
      case MEM_FP:
	return list_mem.at(pos)->r_fpo(val, &trash);
      case MEM_DEST_S:
	return list_mem.at(pos)->r_des(val, &trash);
      case MEM_UNUSED:
        return ERROR_VACIA;
      default:
	return ERROR_TIPO;
    }else{
      if(NO_ERROR == list_mem.at(pos)->r_int(val)) return NO_ERROR;
      if(NO_ERROR == list_mem.at(pos)->r_fpo(val, &trash)) return NO_ERROR;
      return list_mem.at(pos)->r_des(val, &trash);
    }      
}

error_mem Principal::leer_mem(int pos, float *fval){
    if(pos < 0 || pos >= MAXMEM) return ERROR_NOEXISTE;
    if(pos >= p_top && pila_estricta) return ERROR_SUP_TOP;
    return list_mem.at(pos)->r_rea(fval);
}


// Set_top modifica el puntero a cima y su representacion grafica de forma unificada.
// Los parametros que admite son:
//
// IN int n_top -> La posicion del nuevo puntero a cima.
// IN bool p_muestra -> Indicamos a la funcion si queremos que los cambios en el puntero se refleje en la representacion
//                      grafica del programa. Si estamos ejecutando el programa paso a paso debera ser TRUE. En caso de
//                      ejecucion continua deberia ser FALSE.
//
// El metodo pueden leer o escribir en las siguientes variables de clase:
//
//      IN Principal::pila_estricta: Si pila_estricta es TRUE, cualquier dato que pueda haber en posiciones superiores
//                                   al puntero a cima se borrara de forma automatica.
//      OUT Principal::p_top: Se modifica la posicion del puntero a cima.
//      OUT Principal::rep_mem: Si p_muestra y pila_estricta son true, modifica la representacion grafica de la memoria
//                              al modificar el puntero.
//
// Deberiamos comprobar la salida en busca de posibles errores. Los errores pueden ser:
//
//                          NO_ERROR: No se han producido errores. Equivale al valor 0.
//                          ERROR_NOEXISTE: El argumento n_top no esta dentro del rango (0, MAXMEM)

error_mem Principal::set_top(int n_top, bool p_muestra){
    //AMPLIACION: Borrar las posiciones de memoria superiores a TOP cada vez que
    //se decremente el puntero.
    if(n_top < 0 || n_top >= MAXMEM) return ERROR_NOEXISTE;
    if(p_muestra){
	rep_mem.at(p_top)->setIcon(1, QIcon());
	rep_mem.at(n_top)->setIcon(1, QIcon(":/iconos/flecha_azul.png"));
    }
    if(pila_estricta){  
      if(n_top < p_top){
	for(int cont = p_top; cont >= n_top; cont--){
	  if(list_mem.at(cont)->r_tip() != MEM_UNUSED){
	    list_mem.at(cont)->reset();
	    if(p_muestra){
	      rep_mem.at(cont)->setText(3, QString());
	      rep_mem.at(cont)->setText(4, QString());
	      rep_mem.at(cont)->setText(5, QString());
	    } //Fin If
	  } //Fin If
	} //Fin For
      } //Fin If
    } //Fin If
    p_top = n_top;
    return NO_ERROR;
}


// Set_fp modifica el puntero a marco y su representacion grafica de forma unificada.
// Los parametros que admite son:
//
// IN int n_fp -> La posicion del nuevo puntero a marco.
// IN bool p_muestra -> Indicamos a la funcion si queremos que los cambios en el puntero se refleje en la representacion
//                      grafica del programa. Si estamos ejecutando el programa paso a paso debera ser TRUE. En caso de
//                      ejecucion continua deberia ser FALSE.
//
// El metodo pueden leer o escribir en las siguientes variables de clase:
//
//      OUT Principal::p_fp: Se modifica la posicion del puntero a cima.
//      OUT Principal::rep_mem: Si p_muestra es true, modifica la representacion grafica de la memoria
//                              al modificar el puntero (El icono en forma de flecha).
//
// Si la variable p_muestra es true se y el nuevo puntero es distinto al puntero existente se llamara al metodo
// Principal::set_dir_fisicas que cambiara la indicacion de direcciones fisicas de memoria en el panel de instrucciones.
//
// Deberiamos comprobar la salida en busca de posibles errores. Los errores pueden ser:
//
//                          NO_ERROR: No se han producido errores. Equivale al valor 0.
//                          ERROR_NOEXISTE: El argumento n_top no esta dentro del rango (0, MAXMEM)

error_mem Principal::set_fp(int n_fp, bool p_muestra){
    if(n_fp < 0 || n_fp >= MAXMEM) return ERROR_NOEXISTE;
    if(p_muestra){
	rep_mem.at(p_fp)->setIcon(2, QIcon());
	rep_mem.at(n_fp)->setIcon(2, QIcon(":/iconos/flecha_verde.png"));
    }
    bool cambia_dir_fisicas;
    (p_fp != n_fp) ? cambia_dir_fisicas = true : cambia_dir_fisicas = false;
    p_fp = n_fp;
    if(cambia_dir_fisicas && p_muestra) set_dir_fisicas();
    return NO_ERROR;
}
